@@ -17,6 +17,7 @@ this program. If not, see <https://www.gnu.org/licenses/>.
 
 #include "BoardingPanel.h"
 #include "comparators/ByGivenOrder.h"
+#include "CategoryList.h"
 #include "CoreStartData.h"
 #include "Dialog.h"
 #include "text/Font.h"
@@ -28,6 +29,7 @@ this program. If not, see <https://www.gnu.org/licenses/>.
 #include "HailPanel.h"
 #include "LineShader.h"
 #include "MapDetailPanel.h"
+#include "MessageLogPanel.h"
 #include "Messages.h"
 #include "Mission.h"
 #include "Phrase.h"
@@ -36,7 +38,6 @@ this program. If not, see <https://www.gnu.org/licenses/>.
 #include "PlayerInfo.h"
 #include "PlayerInfoPanel.h"
 #include "Preferences.h"
-#include "Random.h"
 #include "Screen.h"
 #include "Ship.h"
 #include "ShipEvent.h"
@@ -81,8 +82,15 @@ void MainPanel::Step()
 		GetUI()->Push(new PlayerInfoPanel(player));
 		isActive = false;
 	}
+	else if(show.Has(Command::MESSAGE_LOG))
+	{
+		GetUI()->Push(new MessageLogPanel());
+		isActive = false;
+	}
 	else if(show.Has(Command::HAIL))
 		isActive = !ShowHailPanel();
+	else if(show.Has(Command::HELP))
+		isActive = !ShowHelp(true);
 	show = Command::NONE;
 
 	// If the player just landed, pop up the planet panel. When it closes, it
@@ -95,39 +103,8 @@ void MainPanel::Step()
 	}
 
 	// Display any relevant help/tutorial messages.
-	const Ship *flagship = player.Flagship();
-	if(flagship)
-	{
-		// Check if any help messages should be shown.
-		if(isActive && flagship->IsTargetable())
-			isActive = !DoHelp("navigation");
-		if(isActive && flagship->IsDestroyed())
-			isActive = !DoHelp("dead");
-		if(isActive && flagship->IsDisabled() && !flagship->IsDestroyed())
-			isActive = !DoHelp("disabled");
-		bool canRefuel = player.GetSystem()->HasFuelFor(*flagship);
-		if(isActive && !flagship->IsHyperspacing() && !flagship->JumpsRemaining() && !canRefuel)
-			isActive = !DoHelp("stranded");
-		shared_ptr<Ship> target = flagship->GetTargetShip();
-		if(isActive && target && target->IsDisabled() && !target->GetGovernment()->IsEnemy())
-			isActive = !DoHelp("friendly disabled");
-		if(isActive && player.Ships().size() > 1)
-			isActive = !DoHelp("multiple ship controls");
-		if(isActive && !flagship->IsHyperspacing() && flagship->Position().Length() > 10000.
-				&& player.GetDate() <= player.StartData().GetDate() + 4)
-		{
-			++lostness;
-			int count = 1 + lostness / 3600;
-			if(count > lostCount && count <= 7)
-			{
-				string message = "lost 1";
-				message.back() += lostCount;
-				++lostCount;
-
-				isActive = !DoHelp(message);
-			}
-		}
-	}
+	if(isActive)
+		isActive = !ShowHelp(false);
 
 	engine.Step(isActive);
 
@@ -216,11 +193,25 @@ bool MainPanel::AllowsFastForward() const noexcept
 
 
 
+Engine &MainPanel::GetEngine()
+{
+	return engine;
+}
+
+
+
 // Only override the ones you need; the default action is to return false.
 bool MainPanel::KeyDown(SDL_Keycode key, Uint16 mod, const Command &command, bool isNewPress)
 {
-	if(command.Has(Command::MAP | Command::INFO | Command::HAIL))
+	if(command.Has(Command::MAP | Command::INFO | Command::MESSAGE_LOG | Command::HAIL | Command::HELP))
 		show = command;
+	else if(command.Has(Command::TURRET_TRACKING))
+	{
+		bool newValue = !Preferences::Has("Turrets focus fire");
+		Preferences::Set("Turrets focus fire", newValue);
+		Messages::Add("Turret tracking mode set to: " + string(newValue ? "focused" : "opportunistic") + ".",
+			Messages::Importance::High);
+	}
 	else if(command.Has(Command::AMMO))
 	{
 		Preferences::ToggleAmmoUsage();
@@ -241,14 +232,6 @@ bool MainPanel::KeyDown(SDL_Keycode key, Uint16 mod, const Command &command, boo
 
 
 
-// Forward the given TestContext to the Engine under MainPanel.
-void MainPanel::SetTestContext(TestContext &testContext)
-{
-	engine.SetTestContext(testContext);
-}
-
-
-
 bool MainPanel::Click(int x, int y, int clicks)
 {
 	// Don't respond to clicks if another panel is active.
@@ -262,8 +245,9 @@ bool MainPanel::Click(int x, int y, int clicks)
 
 	SDL_Keymod mod = SDL_GetModState();
 	hasShift = (mod & KMOD_SHIFT);
+	hasControl = (mod & KMOD_CTRL);
 
-	engine.Click(dragSource, dragSource, hasShift);
+	engine.Click(dragSource, dragSource, hasShift, hasControl);
 
 	return true;
 }
@@ -297,7 +281,7 @@ bool MainPanel::Release(int x, int y)
 	{
 		dragPoint = Point(x, y);
 		if(dragPoint.Distance(dragSource) > 5.)
-			engine.Click(dragSource, dragPoint, hasShift);
+			engine.Click(dragSource, dragPoint, hasShift, hasControl);
 
 		isDragging = false;
 	}
@@ -336,9 +320,7 @@ void MainPanel::ShowScanDialog(const ShipEvent &event)
 					out << "This " + target->Noun() + " is carrying:\n";
 				first = false;
 
-				out << "\t" << it.second
-					<< (it.second == 1 ? " ton of " : " tons of ")
-					<< it.first << "\n";
+				out << "\t" << Format::CargoString(it.second, it.first) << "\n";
 			}
 		for(const auto &it : target->Cargo().Outfits())
 			if(it.second)
@@ -347,14 +329,14 @@ void MainPanel::ShowScanDialog(const ShipEvent &event)
 					out << "This " + target->Noun() + " is carrying:\n";
 				first = false;
 
-				out << "\t" << it.second;
+				out << "\t";
 				if(it.first->Get("installable") < 0.)
 				{
 					int tons = ceil(it.second * it.first->Mass());
-					out << (tons == 1 ? " ton of " : " tons of ") << Format::LowerCase(it.first->PluralName()) << "\n";
+					out << Format::CargoString(tons, Format::LowerCase(it.first->PluralName())) << "\n";
 				}
 				else
-					out << " " << (it.second == 1 ? it.first->Name(): it.first->PluralName()) << "\n";
+					out << it.second << " " << (it.second == 1 ? it.first->DisplayName() : it.first->PluralName()) << "\n";
 			}
 		if(first)
 			out << "This " + target->Noun() + " is not carrying any cargo.\n";
@@ -369,11 +351,14 @@ void MainPanel::ShowScanDialog(const ShipEvent &event)
 			out << "This " + target->Noun() + " is not equipped with any outfits.\n";
 
 		// Split target->Outfits() into categories, then iterate over them in order.
-		auto comparator = ByGivenOrder<string>(GameData::Category(CategoryType::OUTFIT));
+		vector<string> categories;
+		for(const auto &category : GameData::GetCategory(CategoryType::OUTFIT))
+			categories.push_back(category.Name());
+		auto comparator = ByGivenOrder<string>(categories);
 		map<string, map<const string, int>, ByGivenOrder<string>> outfitsByCategory(comparator);
 		for(const auto &it : target->Outfits())
 		{
-			string outfitNameForDisplay = (it.second == 1 ? it.first->Name() : it.first->PluralName());
+			string outfitNameForDisplay = (it.second == 1 ? it.first->DisplayName() : it.first->PluralName());
 			outfitsByCategory[it.first->Category()].emplace(std::move(outfitNameForDisplay), it.second);
 		}
 		for(const auto &it : outfitsByCategory)
@@ -392,7 +377,7 @@ void MainPanel::ShowScanDialog(const ShipEvent &event)
 		for(const Ship::Bay &bay : target->Bays())
 			if(bay.ship)
 			{
-				int &value = count[bay.ship->ModelName()];
+				int &value = count[bay.ship->DisplayModelName()];
 				if(value)
 				{
 					// If the name and the plural name are the same string, just
@@ -439,7 +424,7 @@ bool MainPanel::ShowHailPanel()
 
 	if(flagship->IsEnteringHyperspace())
 		Messages::Add("Unable to send hail: your flagship is entering hyperspace.", Messages::Importance::High);
-	else if(flagship->Cloaking() == 1.)
+	else if(flagship->IsCloaked() && !flagship->Attributes().Get("cloaked communication"))
 		Messages::Add("Unable to send hail: your flagship is cloaked.", Messages::Importance::High);
 	else if(target)
 	{
@@ -447,7 +432,7 @@ bool MainPanel::ShowHailPanel()
 		// because the player has no way of telling if it's presently jumping or
 		// not. If it's in system and jumping, report that.
 		if(target->Zoom() < 1. || target->IsDestroyed() || target->GetSystem() != player.GetSystem()
-				|| target->Cloaking() == 1.)
+				|| target->IsCloaked())
 			Messages::Add("Unable to hail target " + target->Noun() + ".", Messages::Importance::High);
 		else if(target->IsEnteringHyperspace())
 			Messages::Add("Unable to send hail: " + target->Noun() + " is entering hyperspace."
@@ -486,6 +471,129 @@ bool MainPanel::ShowHailPanel()
 
 
 
+bool MainPanel::ShowHelp(bool force)
+{
+	const Ship *flagship = player.Flagship();
+	if(!flagship)
+		return false;
+
+	vector<string> forced;
+	// Check if any help messages should be shown.
+	if(Preferences::Has("Control ship with mouse"))
+	{
+		if(force)
+			forced.push_back("control ship with mouse");
+		else if(DoHelp("control ship with mouse"))
+			return true;
+	}
+	if(flagship->IsTargetable())
+	{
+		if(force)
+			forced.push_back("navigation");
+		else if(DoHelp("navigation"))
+			return true;
+	}
+	if(flagship->IsDestroyed())
+	{
+		if(force)
+			forced.push_back("dead");
+		else if(DoHelp("dead"))
+			return true;
+	}
+	else if(flagship->IsDisabled())
+	{
+		if(force)
+			forced.push_back("disabled");
+		else if(DoHelp("disabled"))
+			return true;
+	}
+	bool canRefuel = player.GetSystem()->HasFuelFor(*flagship);
+	if(!flagship->IsHyperspacing() && !flagship->JumpsRemaining() && !canRefuel)
+	{
+		if(force)
+			forced.push_back("stranded");
+		else if(DoHelp("stranded"))
+			return true;
+	}
+	shared_ptr<Ship> target = flagship->GetTargetShip();
+	if(target && target->IsDisabled() && !target->GetGovernment()->IsEnemy())
+	{
+		if(force)
+			forced.push_back("friendly disabled");
+		else if(DoHelp("friendly disabled"))
+			return true;
+	}
+	if(player.Ships().size() > 1)
+	{
+		if(force)
+			forced.push_back("multiple ship controls");
+		else if(DoHelp("multiple ship controls"))
+			return true;
+	}
+	if(flagship->IsTargetable() && player.Ships().size() > 1)
+	{
+		if(force)
+			forced.push_back("fleet harvest tutorial");
+		else if(DoHelp("fleet harvest tutorial"))
+			return true;
+	}
+	if(flagship->IsTargetable() &&
+			flagship->Attributes().Get("asteroid scan power") &&
+			player.Ships().size() > 1)
+	{
+		// Different order of these messages is intentional,
+		// because we're displaying the forced messages in reverse order.
+		if(force)
+		{
+			forced.push_back("fleet asteroid mining");
+			forced.push_back("fleet asteroid mining shortcuts");
+		}
+		else if(DoHelp("fleet asteroid mining shortcuts") && DoHelp("fleet asteroid mining"))
+			return true;
+	}
+	if(player.DisplayCarrierHelp())
+	{
+		if(force)
+			forced.push_back("try out fighters transfer cargo");
+		else if(DoHelp("try out fighters transfer cargo"))
+			return true;
+	}
+	if(Preferences::Has("Fighters transfer cargo"))
+	{
+		if(force)
+			forced.push_back("fighters transfer cargo");
+		else if(DoHelp("fighters transfer cargo"))
+			return true;
+	}
+	if(!flagship->IsHyperspacing() && flagship->Position().Length() > 10000.
+			&& player.GetDate() <= player.StartData().GetDate() + 4)
+	{
+		++lostness;
+		int count = 1 + lostness / 3600;
+		if(count > lostCount && count <= 7)
+		{
+			string message = "lost 1";
+			message.back() += lostCount;
+			++lostCount;
+			if(force)
+				forced.push_back(message);
+			else if(DoHelp(message))
+				return true;
+		}
+	}
+
+	if(!force || forced.empty())
+		return false;
+
+	bool hasValidHelp = false;
+	// Reverse-iterate so that the player will see the basic messages first.
+	for(auto it = forced.rbegin(); it != forced.rend(); ++it)
+		hasValidHelp |= DoHelp(*it, true);
+	return hasValidHelp;
+}
+
+
+
 // Handle ShipEvents from this and previous Engine::Step calls. Start with the
 // oldest and then process events until any create a new UI element.
 void MainPanel::StepEvents(bool &isActive)
