@@ -21,7 +21,9 @@ this program. If not, see <https://www.gnu.org/licenses/>.
 #include "Angle.h"
 #include "Point.h"
 
+#include <cstdint>
 #include <memory>
+#include <set>
 #include <vector>
 
 class Government;
@@ -50,10 +52,6 @@ public:
 	};
 
 
-public:
-	static Angle Inaccuracy(double value);
-
-
 public:
 	Projectile(const Ship &parent, Point position, Angle angle, const Weapon *weapon);
 	Projectile(const Projectile &parent, const Point &offset, const Angle &angle, const Weapon *weapon);
@@ -71,10 +69,12 @@ public:
 	// Move the projectile. It may create effects or submunitions.
 	void Move(std::vector<Visual> &visuals, std::vector<Projectile> &projectiles);
 	// This projectile hit something. Create the explosion, if any. This also
-	// marks the projectile as needing deletion.
+	// marks the projectile as needing deletion if it has run out of penetrations.
 	void Explode(std::vector<Visual> &visuals, double intersection, Point hitVelocity = Point());
 	// Get the amount of clipping that should be applied when drawing this projectile.
 	double Clip() const;
+	// Get whether the lifetime on this projectile has run out.
+	bool IsDead() const;
 	// This projectile was killed, e.g. by an anti-missile system.
 	void Kill();
 
@@ -84,7 +84,7 @@ public:
 	// Get information on the weapon that fired this projectile.
 	const Weapon &GetWeapon() const;
 	// Get information on how this projectile impacted a ship.
-	ImpactInfo GetInfo() const;
+	ImpactInfo GetInfo(double intersection) const;
 
 	// Find out which ship or government this projectile is targeting. Note:
 	// this pointer is not guaranteed to be dereferenceable, so only use it
@@ -99,6 +99,16 @@ public:
 
 	// Get the distance that this projectile has traveled.
 	double DistanceTraveled() const;
+	// Get the number of objects this projectile can still collide with.
+	uint16_t HitsRemaining() const;
+	// Get whether this projectile should explode the next time collision
+	// detection is run.
+	bool ShouldExplode() const;
+
+	// Once the projectile has come into contact with a phasing device, it
+	// will be decided if it should completely phase through or make contact.
+	bool Phases(const Ship &ship) const;
+	void SetPhases(const Ship *ship);
 
 
 private:
@@ -110,6 +120,7 @@ private:
 
 	std::weak_ptr<Ship> targetShip;
 	const Ship *cachedTarget = nullptr;
+	bool targetDisabled = false;
 	const Government *targetGovernment = nullptr;
 
 	// The change in velocity of all stages of this projectile
@@ -117,8 +128,13 @@ private:
 	Point dV;
 	double clip = 1.;
 	int lifetime = 0;
-	double distanceTraveled = 0;
+	double distanceTraveled = 0.;
+	uint16_t hitsRemaining = 1U;
 	bool hasLock = true;
+
+	// This is safe to keep even if the ships die, because we don't actually call the ship,
+	// we just compare this pointer to other ship pointers.
+	const Ship *phasedShip;
 };
 
 
